
#include <sys/types.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/dir.h>
#include <regex.h>
#include "stgp.h"
/* tried to redefine the tokens that lemon uses to custom ones 
 * and it will not work. It appears that the positions of the 
 * tokens is matched with their numbers */
#include "ibp.h"
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/epoll.h>

#define NUMACCTS 2
#define TIMEOUT 1000 /* in milliseconds */
#define MAXINACTIVETIME 600 /* in seconds */
#define DEBUG 5
#define TEST 1
#define FILE_MODE (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)
#define DIR_MODE (FILE_MODE | S_IXUSR | S_IXGRP | S_IXOTH)
#define MAXCHILDREN 10 
#define READBUFSIZE 8192
#define DB(error_level,args) if (error_level <= DEBUG)  {printf("%s:%4d:", __FILE__,__LINE__), indent(error_level),printf("Begin "), printf(args), printf("\n");}
#define DP(error_level,args) if (error_level <= DEBUG) {printf("%s:%4d:",__FILE__,__LINE__), indent(error_level), printf args, printf("\n");}
#define DE(error_level,args) if (error_level <= DEBUG)  {printf("%s:%4d:", __FILE__,__LINE__), indent(error_level), printf("End "), printf(args), printf("\n");}
#undef TEST 
#ifdef TEST
  #define TI(variable_initializations)  variable_initializations 
  #define TCODE( test_code )  test_code
    /* need a better printing statement here */
  #define TC(cond,stmt)    {\
            if (!(cond)) {\
	                  printf ("%4d:test failed for: %s with ", __LINE__, ( # cond ) );\
	                  printf stmt;\
	                  printf("\n");\
	                  exit(1);\
	                  }\
          }
#else
  #define TI( variable_initializations )  ;
  #define TCODE( test_code )  ;
  #define TC(cond, var)   ;
#endif


/* re2c stuff from here */

/* typedef unsigned int uint; */ /* defined in /usr/include/sys/types.h */
typedef unsigned char uchar;
typedef struct Scanner {
    int			fd;
    uint		line;
    uchar		*bot, *tok, *ptr, *cur, *pos, *lim, *top, *eof;
} Scanner;

uchar * fill(Scanner *s, uchar *cursor);
int ibl_value_scan(Scanner *s);
int ibl_token_scan(Scanner *s);
int stgscan(Scanner *s);

#define	BSIZE	8192

#define	YYCTYPE		uchar
#define	YYCURSOR	cursor
#define	YYLIMIT		s->lim
#define	YYMARKER	s->ptr
#define	YYFILL(n)	{cursor = fill(s, cursor);}

#define	RET(i)	{s->cur = cursor; return i;}

struct trade {
   int index;
   int orderid;
   int quantity;
   int parent;
   int children[MAXCHILDREN];
   int no_of_children;
   int status;
   double trade_price;
   char trade_operator[3];
   double trigger_price;
   char trigger_operator[3];
   double trailing_margin;
   char trailing_operator[3];
   char price_type[5];
   int pending_execution;
};

struct stock {
   int current_trade_state;
   char symbol[10];
   char localsymbol[10];
   char sectype[5];
   char expiry[10];
   char currency[5];
   char exchange[7];
   char right[5];
   double strike;
   double last;
   double ask;
   double bid;
   int roottrade;
   int requested_market_data;
   char multiplier[6];
};
#define WRITEBUFSIZE 256
struct data {
  int request_contract_details;
  int request_market_data;
  time_t time_data_recvd;
  int bytes_recvd;
  int clientid;
  int orderid;
  int ssize;
  int sempty;
  time_t stks_mtime;
  struct stock * stks;
  int tsize;
  int tempty;
  struct trade * trds;
  char account[10];
  int port;
  int sockfd;
  int cptrn;
  int cstk; /* a counter */
  char rbuf[READBUFSIZE];
  char wbuf[WRITEBUFSIZE];
  int startloc;
  void * parser;
  struct Scanner scanner;
  int close_connection;
  int write_current_state;
  int pending_orders;
};

struct pdata {
  int psize;
  int pempty; 
  struct pattern * ptrns;
/*  time_t st_mtime; */
  time_t mtime;
};

#define NAMESIZE 40
struct field {
  int position;
  int type; /* 0 int 1 float 2 string */
  char name[NAMESIZE];
  char field[NAMESIZE];
};

struct pattern {
  /* char *pattern; 
  regex_t rgx;
  regmatch_t * mtchs;
  int ptsize; */
  char ptname[NAMESIZE];
  int fsize;
  int fempty;
  struct field * flds;
};

#define TOKVALUESIZE 200
 /* hardcoded in main.h There is a problem with addresses or pointers for token value, as by the time the parser reads it off the stack the address is pointing to a different value */

struct tokenvalue {
  /* int	idx; */
   char value[TOKVALUESIZE];
};

struct stgptokenvalue {
   int	actidx; 
   char value[TOKVALUESIZE];
};

struct stgpnonterminal {
  int actidx;
  int trdidx;
  int stkidx;
 };



/* these function declarations are being added here to avoid these:
 * warning: implicit declaration of function `ParseAlloc' */

void *stgparseAlloc(void *(*mallocProc)(size_t));
void stgparseFree(
      void *p,                    /* The parser to be deleted */
        void (*freeProc)(void*)     /* Function used to reclaim memory */
    );
void stgparse(
      void *yyp,                   /* The parser */
      int yymajor,                 /* The major token code number */
      struct stgptokenvalue yyminor,    /* The value for the token */
      struct data * dataptr   /* Optional %extra_argument parameter */
    );

void *ibparseAlloc(void *(*mallocProc)(size_t));
void ibparseFree(
      void *p,                    /* The parser to be deleted */
        void (*freeProc)(void*)     /* Function used to reclaim memory */
    );
void ibparse(
      void *yyp,                   /* The parser */
      int yymajor,                 /* The major token code number */
      struct tokenvalue yyminor,    /* The value for the token */
      struct data * dataptr   /* Optional %extra_argument parameter */
    );

int ibparse_next_action_is_a_shift(
      void *pParser        /* The parser */
    );

/* from common.c */
void indent(int level);
int increase (void ** pointer, int size, int element_size);

void write_strings(int sockfd, char * buf, int count, ...);
void write_integers(int sockfd, char * buf, int count, ...);
void write_floats(int sockfd, char * buf, int count, ...);

int next_fld (struct pattern * patternptr );
int next_stk (struct data * dataptr );
int next_trd (struct data * dataptr );

void printtree(int depth, struct data * dataptr, int trdindex,FILE * fd);
void compare_update(struct data * stocks, struct data * currstocks );
int get_latest_file(char * latest_filename, char * subdir, time_t * st_mtimeptr);
void get_strategy( struct data * dataptr, char * dir, time_t * mtimeptr);
void  write_current_state(struct data * dataptr);
void email_current_state ();
void send_close_window_message ();
void read_strategy( struct data * dataptr);


/* gcc -oO3 -ggdb -Wall ib.c -o ib */

	#define CLIENT_VERSION 18
	#define EOL 0
	#define SERVER_VERSION 14

	#define REQ_MKT_DATA 1
	#define CANCEL_MKT_DATA 2
	#define PLACE_ORDER 3
	#define CANCEL_ORDER 4
	#define REQ_OPEN_ORDERS 5
	#define REQ_ACCOUNT_DATA 6
	#define REQ_EXECUTIONS 7
	#define REQ_IDS 8
	#define REQ_CONTRACT_DATA 9
	#define REQ_MKT_DEPTH 10
	#define CANCEL_MKT_DEPTH 11
	#define REQ_NEWS_BULLETINS 12
	#define CANCEL_NEWS_BULLETINS 13
	#define SET_SERVER_LOG_LEVEL 14
	#define REQ_AUTO_OPEN_ORDERS 15
	#define REQ_ALL_OPEN_ORDERS 16
	#define REQ_MANAGED_ACCTS 17
	#define REQ_FA 18
	#define REPLACE_FA 19

	#define NEGATIVE_MSG -1
	#define TICK_PRICE 1
	#define TICK_SIZE 2
	#define ORDER_STATUS 3
	#define ERR_MSG 4
	#define OPEN_ORDER 5
	#define ACCT_VALUE 6
	#define PORTFOLIO_VALUE 7
	#define ACCT_UPDATE_TIME 8
	#define NEXT_VALID_ID 9 
	#define CONTRACT_DATA 10
	#define EXECUTION_DATA 11
	#define MARKET_DEPTH 12
	#define MARKET_DEPTH_L2 13
	#define NEWS_BULLETINS 14
	#define MANAGED_ACCTS 15
	#define RECEIVE_FA 16

	enum possible_status {NONE, OPEN, TRIGGERED, TRAILING, PLACEORDER, TWS_OPEN_ORDER, PARTIALLY_CLOSED, CLOSED};
/* 
	const char * status_descr[] = {"NONE", "OPEN", "TRIGGERED", "TRAILING", "PLACE_ORDER", "CLOSED"};
	*/


/* gcc -oO3 -ggdb -Wall ib.c -o ib  */


#define MARKET_START_TIME "14:35" /* The current system time is 3 mins faster than the time displayed on I */
#define MARKET_END_TIME "21:00"

#define TRUE 1
#define FALSE 0
#define SERVER_LOG_LEVEL 3

/* from ib.c */
void close_connection(struct data * dataptr,struct epoll_event * recvd_events_ptr);
void read_data_from_socket(struct data * dataptr,struct epoll_event * recvd_events_ptr);
int request_market_data ( struct data * dataptr );
int request_contract_details( struct data * dataptr );
int connect_to_ib (struct data * dataptr, int * epollfdptr, struct epoll_event * event_to_registerptr, struct epoll_event * recvd_events_ptr);
void parse_data( struct data * dataptr );
int strategy_state_change ( struct data * dataptr );
int place_order ( struct data * dataptr);
void assign_tick_price ( struct data * dataptr, int tickerid, int type, float price) ;
void change_order_status ( struct data * dataptr, int status,
      			int orderId, char * symbol,char * secType, char * expiry,
                	double strike, char * right,char * exchange, char * currency,
                	char * localsymbol, char * action, int totalquantity);


/* testing from here  */
int test_ibparser(struct data * dataptr);
